/*
 *      PIC16F84 controller for G0MRF's DDS longwave exciter
 *
 *      This version is written for 136 kHz F/4 mode only!
 *
 *      Version 1.0 2000-03-02
 *
 *      Written by Johan Bodin, SM6LKM.
 *
 *      Compiler: CC5X free edition version 3.0G
 *      (available at http://www.bknd.com)
 *
 *      The #included files, 16f84.h and int16cxx.h,
 *      are those supplied in the compiler package.
 */

#include <16f84.h>             // processor definitions
#include <int16cxx.h>          // interrupt save/restore macros


/*
 *      Set configuration word
 *
 *      Bits (PIC16F84):
 *        C = CP     (0 = code memory protected)
 *        D = DP     (0 = data EEPROM protected)
 *        P = /PWRTE (0 = power-up timer enabled)
 *        W = WDTE   (0 = watchdog timer disabled)
 *        XY = FOSC1 FOSC0 (00, 01, 10, 11 = LP, XT, HS, RC)
 */
                 // CCCCCCDCCCPWXY
#pragma config |= 0b11111111110001


/*
 *      I/O stuff
 */

#define o_sclk   RA2
#define o_sdata  RA3
// RA4 is currently unused (connected to AD9832 FSEL + pull-up)
#define i_dfcw   RB2  // DFCW offset added when pin is pulled LOW
#define i_lock   RB3
#define o_lcd_rs RB4  // also DFCW pushbutton input (to gnd via 820 ohms)
#define o_lcd_rw RB5  // also CAL pushbutton input (to gnd via 820 ohms)
#define o_lcd_e  RB6
#define o_fsync  RB7


/*
 *      Some funny #defines
 */

#define FALSE   0       // boolean constants
#define TRUE    1


/*
 *      Load default data into EEPROM
 *
 *      Note:
 *        This data is loaded into the EEPROM when the
 *        PIC chip is programmed (no code is generated
 *        by this section).
 *
 *        Quantities larger than 8 bits are stored with LSByte first.
 */

#define EEBASE 0x2100  // EEPROM pseudo base address for hex file
#define eeOSC  0x00    // EEPROM offset for 32-bit oscillator constant

#pragma cdata[EEBASE+eeOSC+0] = 0x12   // 0xABCC7712 = 2^56 / 25 MHz
#pragma cdata[EEBASE+eeOSC+1] = 0x77
#pragma cdata[EEBASE+eeOSC+2] = 0xCC
#pragma cdata[EEBASE+eeOSC+3] = 0xAB


/*
 *      Global variables
 */

char osc0 @ 0x0C;       // oscillator constant at start of free RAM
char osc1 @ 0x0D;
char osc2 @ 0x0E;
char osc3 @ 0x0F;

char min0, min1,        // 32-bit minimum value for knob
     min2, min3;

char max0, max1,        // 32-bit maximum value for knob
     max2, max3;

char base_fq0,          // 32-bit DDS frequency in Hz (without DFCW offset!)
     base_fq1,
     base_fq2,
     base_fq3;

char fq0, fq1,          // 32-bit actual DDS frequency (with DFCW offset)
     fq2, fq3;

char knob0, knob1,      // 32-bit up/down counter for knob encoder
     knob2, knob3;

char encstat;           // encoder change-of-state detector
char bDFWC_Offset;      // DDS DFCW offset in 1 Hz increments
uns16 wStepSize;        // DDS frequency step in 1 Hz increments
bit qKnobChange;        // knob has moved (flag from ISR to main loop)
bit qDFCW_Switch;       // copy of switch level at RB4 (muxed I/O on this pin)
char bDeb_RB5;          // debounce counter for CAL switch at RB5 (also muxed)
bit qCalEdge;           // 1->0 transition on CAL switch (debounced)
bit qLastDFCW;          // transition detector for i_dfcw input pin
bit qNewFreq;
bit qLastLock;


/*
 *      Interrupt service routine
 *
 *      This is a "tick" routine that is driven by the TMR0 interrupt.
 *      Execution rate is 3906.25 Hz with 4 MHz crystal.
 */

#define KNOBRATE 10
#define KNOBMAX  256

#pragma origin 4

interrupt vIntServer (void)
{
  char bPCLATH_Save;
  static char bSlowDown;

  int_save_registers        // save W and status
  bPCLATH_Save = PCLATH;    // save PCLATH

  T0IF = 0;                 // clear interrupt flag

  //    "Automatic gearbox" for encoder knob

  if (--bSlowDown == 0)
  {
    bSlowDown = KNOBRATE;   // number of ticks between halving step size
    wStepSize >>= 1;        // KNOBRATE ticks has elapsed, reduce step size
    if (wStepSize == 0)
      wStepSize = 1;        // oops! step became zero, must be at least one!
  }

  encstat = rl (encstat);   // shift encstat two bits left to get previous...
  encstat = rl (encstat);   // ...encoder sample in bits 3..2
  encstat &= 0b00001100;    // strip garbage other than bits 3..2
  encstat |= PORTB & 0b00000011;  // put current encoder sample in bits 1..0
  if (!i_lock)              // dial unlocked?
  {
    skip (encstat);
      goto enc_done;        // 00 -> 00 = no change
      goto enc_inc;         // 00 -> 01 = increment
      goto enc_dec;         // 00 -> 10 = decrement
      goto enc_done;        // 00 -> 11 = illegal, ignore
      goto enc_dec;         // 01 -> 00 = decrement
      goto enc_done;        // 01 -> 01 = no change
      goto enc_done;        // 01 -> 10 = illegal, ignore
      goto enc_inc;         // 01 -> 11 = increment
      goto enc_inc;         // 10 -> 00 = increment
      goto enc_done;        // 10 -> 01 = illegal, ignore
      goto enc_done;        // 10 -> 10 = no change
      goto enc_dec;         // 10 -> 11 = decrement
      goto enc_done;        // 11 -> 00 = illegal, ignore
      goto enc_dec;         // 11 -> 01 = decrement
      goto enc_inc;         // 11 -> 10 = increment
      goto enc_done;        // 11 -> 11 = no change

#pragma computedGoto 0      // end computed goto region

enc_dec:
#asm
        bsf     qKnobChange     ; tell main loop that knob has been changed

;       knob -= wStepSize

        movf    wStepSize+0,w
        subwf   knob0,f
        movf    wStepSize+1,w
        btfss   Carry
        incfsz  wStepSize+1,w
        subwf   knob1,f
        movlw   1
        btfsc   Carry
        goto    dec_done
        subwf   knob2,f
        btfsc   Carry
        goto    dec_done
        subwf   knob3,f
dec_done:
#endasm
  wStepSize = KNOBMAX;      // knob is rotated, throttle up tuning speed
#asm

;       if ((int32)knob < (int32)min)
;         knob = min;

        movf    min3,w
        subwf   knob3,w
        btfsc   Zero_
        goto    check_low_bytes
        movf    knob3,w
        xorwf   min3,w
        btfsc   Carry
        xorlw   0x80
        andlw   0x80
        btfsc   Zero_
        goto    setmin
        goto    enc_done
check_low_bytes:
        movf    min2,w
        subwf   knob2,w
        btfss   Carry
        goto    setmin
        btfss   Zero_
        goto    enc_done
        movf    min1,w
        subwf   knob1,w
        btfss   Carry
        goto    setmin
        btfss   Zero_
        goto    enc_done
        movf    min0,w
        subwf   knob0,w
        btfsc   Carry
        goto    enc_done

setmin:
        movf    min0,w
        movwf   knob0
        movf    min1,w
        movwf   knob1
        movf    min2,w
        movwf   knob2
        movf    min3,w
        movwf   knob3
        goto    enc_done
#endasm

enc_inc:
#asm
        bsf     qKnobChange     ; tell main loop that knob has been changed

;       knob += wStepSize

        movf    wStepSize+0,w
        addwf   knob0,f
        movf    wStepSize+1,w
        btfsc   Carry
        incfsz  wStepSize+1,w
        addwf   knob1,f
        btfss   Carry
        goto    inc_done
        incfsz  knob2,f
        goto    inc_done
        incf    knob3,f
inc_done:
#endasm
  wStepSize = KNOBMAX;      // knob is rotated, throttle up tuning speed
#asm

;       if ((uns32)knob > (uns32)max)
;         knob = max;

        movf    knob3,w
        subwf   max3,w
        btfss   Carry
        goto    setmax
        btfss   Zero_
        goto    enc_done
        movf    knob2,w
        subwf   max2,w
        btfss   Carry
        goto    setmax
        btfss   Zero_
        goto    enc_done
        movf    knob1,w
        subwf   max1,w
        btfss   Carry
        goto    setmax
        btfss   Zero_
        goto    enc_done
        movf    knob0,w
        subwf   max0,w
        btfsc   Carry
        goto    enc_done

setmax:
        movf    max0,w
        movwf   knob0
        movf    max1,w
        movwf   knob1
        movf    max2,w
        movwf   knob2
        movf    max3,w
        movwf   knob3
#endasm
  }  // end of "if dial unlocked"

enc_done:
  PCLATH = bPCLATH_Save;    // restore PCLATH
  int_restore_registers     // restore W and status
}


/*
 *      Send a serial 16-bit word to the DDS chip
 *
 *      The argument wData is sent to the DDS chip, MSB first.
 */

void vSend_DDS (uns16 wData)
{
  char bLoop = 16;        // 16 bits to transmit

  o_fsync = 0;            // AD9832 FSYNC pin low
  do
  {
    o_sdata = wData.15;   // shift out data MSBit first
    o_sclk = 1;           // make clock pulse on AD9832 SCLK pin
    o_sclk = 0;
    wData = rl (wData);   // shift data bits to the left
  } while (--bLoop);      // all 16 bits transmitted?
  o_fsync = 1;            // AD9832 FSYNC pin high
}


/*
 *      Compute phase increment from frequency and send to DDS chip
 *
 *      Input:
 *        fq3:fq2:fq1:fq0 = 32 bit DDS frequency in Hz
 *        osc3:osc2:osc1:osc0 = 32 bit oscillator constant = 2^56/fosc
 *
 *      Output:
 *        The DDS chip will be updated with the new frequency.
 */

void vUpdate_DDS (void)
{
  static bit qRegToggle;
  char bLoop;
  char osc3temp, osc2temp, osc1temp, osc0temp;
  char dds4, dds3, dds2, dds1, dds0;  // dds4 needed in calculation only

  osc0temp = osc0;  // don't mess with original value, use a copy instead
  osc1temp = osc1;
  osc2temp = osc2;
  osc3temp = osc3;

  dds0 = 0;         // clear result
  dds1 = 0;
  dds2 = 0;
  dds3 = 0;
  dds4 = 0;
  bLoop = 32;       // 32 bits to process
#asm
multloop:
        bcf     Carry
        btfss   osc0temp,0      // add this binary weight?
        goto    shiftit

        movf    fq0,w           // dds += ((uns40)fq << 8)
        addwf   dds1,f
        movf    fq1,w
        btfsc   Carry
        incfsz  fq1,w
        addwf   dds2,f
        movf    fq2,w
        btfsc   Carry
        incfsz  fq2,w
        addwf   dds3,f
        movf    fq3,w
        btfsc   Carry
        incfsz  fq3,w
        addwf   dds4,f
shiftit:
        rrf     dds4,f          // rotate it all
        rrf     dds3,f
        rrf     dds2,f
        rrf     dds1,f
        rrf     dds0,f
        rrf     osc3temp,f
        rrf     osc2temp,f
        rrf     osc1temp,f
        rrf     osc0temp,f
        decfsz  bLoop,f
        goto    multloop
#endasm

 // Update DDS chip with a new frequency word (dds4 is not part of result)

  if (qRegToggle)  // load data into frequency register 0
  {
    qRegToggle = 0;
    vSend_DDS (0x3000 | dds0);  // write to defer register, destination = 0
    vSend_DDS (0x2100 | dds1);  // write 16 frequency bits, destination = 1
    vSend_DDS (0x3200 | dds2);  // write to defer register, destination = 2
    vSend_DDS (0x2300 | dds3);  // write 16 frequency bits, destination = 3
    vSend_DDS (0x5000);         // switch to frequency register 0
    return;
  }
  else             // load data into frequency register 1
  {
    qRegToggle = 1;
    vSend_DDS (0x3400 | dds0);  // write to defer register, destination = 4
    vSend_DDS (0x2500 | dds1);  // write 16 frequency bits, destination = 5
    vSend_DDS (0x3600 | dds2);  // write to defer register, destination = 6
    vSend_DDS (0x2700 | dds3);  // write 16 frequency bits, destination = 7
    vSend_DDS (0x5800);         // switch to frequency register 1
  }
}


/*
 *      Delay bTime*1280+4 processor cycles (including call & return)
 *
 *      Note: caller's W is bombed!
 */

void vDelay (char bTime)
{
#asm
        clrw
waste_more_time:
        addlw   1
        btfsc   Zero_
        decfsz  bTime,f
        goto    waste_more_time
#endasm
}


/*
 *      Wait for LCD to be ready for new data
 */

void vLCD_BusyWait (void)
{
  TRISA = 0x1F;         // tristate RA0..RA3 (RA4 is also input!)
  o_lcd_rs = 0;         // select LCD command register
  o_lcd_rw = 1;         // setup LCD for read

  do
  {
    o_lcd_e = 1;        // E high
    W = PORTA;          // read LCD D7..D4 into W bits 3..0
    o_lcd_e = 0;        // E low
    o_lcd_e = 1;        // another E pulse to dummy read LCD D3..D0
    o_lcd_e = 0;
  } while (W & 0x08);   // LCD is busy if D7 was set (W bit 3)

  o_lcd_rw = 0;         // setup LCD for write
  TRISA = 0x10;         // make RA0..RA3 outputs again
}


/*
 *      Send a byte to LCD control register
 */

void vLCD_Ctrl (char bCtrl)
{
  vLCD_BusyWait ();      // wait for LCD to be ready
  o_lcd_rs = 0;          // select LCD command register
  PORTA = swap (bCtrl);  // write bits 7..4
  o_lcd_e = 1;           // make E pulse
  o_lcd_e = 0;
  PORTA = bCtrl;         // write bits 3..0
  o_lcd_e = 1;           // make E pulse
  o_lcd_e = 0;
}


/*
 *      Send a byte to LCD data register
 */

void vLCD_Data (char bData)
{
  vLCD_BusyWait ();      // wait for LCD to be ready
  o_lcd_rs = 1;          // select LCD data register
  PORTA = swap (bData);  // write bits 7..4
  o_lcd_e = 1;           // make E pulse
  o_lcd_e = 0;
  PORTA = bData;         // write bits 3..0
  o_lcd_e = 1;           // make E pulse
  o_lcd_e = 0;
}


/*
 *      Show fractional frequency in quarters of Hz
 *
 *      Argument W = 0..3 results in .00 to .75 beeing
 *      written to the LCD at the current cursor position.
 */

void vShowFraction (char frac)
{
  char ascii = 0x30;

  vLCD_Data ('.');    // write decimal point
  if (frac.1)
    ascii += 5;
  if (frac.0)
    ascii.1 = 1;
  vLCD_Data (ascii);  // write '0', '2', '5' or '7'
  if (frac.0)
    vLCD_Data ('5');  // write '0' or '5'
  else
    vLCD_Data ('0');
}


/*
 *      Write a byte as hex at current cursor position
 */

void vWriteHexByte (char data)
{
  char b;

  b = data >> 4;
  if (b > 9)
    b += 55;
  else
    b += 48;
  vLCD_Data (b);

  b = data & 0x0F;
  if (b > 9)
    b += 55;
  else
    b += 48;
  vLCD_Data (b);
}


/*
 *      Divide the fq by 4, convert the result to BCD and write
 *      as ASCII to the LCD at the current cursor position.
 *
 *      Display format: "137.705.25"
 *
 *      The data in fq3..fq0 is DESTROYED!
 */

// First a little sub' that was impossible to embed in the function
// vShowFrequency() because the compiler refuses absolute asm CALL's...

void adjbcd (void)
{
  char bcdtmp;
#asm
        movlw   0x03
        addwf   INDF,w
        movwf   bcdtmp
        btfsc   bcdtmp,3
        movwf   INDF
        movlw   0x30
        addwf   INDF,w
        movwf   bcdtmp
        btfsc   bcdtmp,7
        movwf   INDF
#endasm
}

void vShowFrequency (void)
{
  char fraction, bLoop;
  char bcd0, bcd1, bcd2, bcd3, bcd4;

  fraction = fq0;    // save LSB before division, used for fractional Hz

  fq3 = rr (fq3);               // fq >>= 2, divide by 4
  fq2 = rr (fq2);
  fq1 = rr (fq1);
  fq0 = rr (fq0);
  fq3 = rr (fq3);
  fq2 = rr (fq2);
  fq1 = rr (fq1);
  fq0 = rr (fq0);
  fq3 &= 0x3F;                  // clear garbage shifted in

  bLoop = 32;                   // 32 bits to convert
  bcd0 = 0;
  bcd1 = 0;
  bcd2 = 0;
  bcd3 = 0;
  bcd4 = 0;

bcdloop:
  fq0 = rl (fq0);
  fq1 = rl (fq1);
  fq2 = rl (fq2);
  fq3 = rl (fq3);

  bcd0 = rl (bcd0);
  bcd1 = rl (bcd1);
  bcd2 = rl (bcd2);
  bcd3 = rl (bcd3);
  bcd4 = rl (bcd4);

#asm
        decfsz  bLoop,f
        goto    adjust
#endasm

  vLCD_Data ((bcd2 >> 4) + 0x30);       // 100 kHz digit
  vLCD_Data ((bcd2 & 0x0F) + 0x30);     // 10 kHz digit
  vLCD_Data ((bcd1 >> 4) + 0x30);       // 1 kHz digit
  vLCD_Data ('.');                      // decimal point
  vLCD_Data ((bcd1 & 0x0F) + 0x30);     // 100 Hz digit
  vLCD_Data ((bcd0 >> 4) + 0x30);       // 10 Hz digit
  vLCD_Data ((bcd0 & 0x0F) + 0x30);     // 1 Hz digit
  vShowFraction (fraction);  // show .00 .25 .50 or .75 (bits 1..0 used)
  return;

adjust:
  FSR = &bcd0;
  adjbcd ();
  FSR++;
  adjbcd ();
  FSR++;
  adjbcd ();
  FSR++;
  adjbcd ();
  FSR++;
  adjbcd ();
  goto bcdloop;
}


/*
 *      Show DFCW offset at current cursor position
 *
 *      0..200 (Hz) is displayed as " 0.00".."50.00"
 */

void vShow_DFCW_Offs (char b)
{
  char ascii, fraction;

  fraction = b;                // save fractional Hz
  b >>= 2;                     // b = integral part of f/4

  ascii = '0';
  while (b >= 10)              // count tens
  {
    ascii++;
    b -= 10;                   // remove all 10's counted
  }
  if (ascii == '0')            // replace leading zero with space
    ascii = 32;
  vLCD_Data (ascii);           // show 10's digit (or space)
  vLCD_Data ('0' + b);         // show 1's digit (that's what's left in b)
  vShowFraction (fraction);    // show .00 .25 .50 or .75 (bits 1..0 used)
}


/*
 *      Write a byte to EEPROM
 *
 *      Caller should first load the appropriate write address
 *      into EEADR (PIC register), then call vEE_Write() with
 *      the data to program as argument.
 *
 *      EEADR will be incremented by one when programming is finished.
 *      Subsequent write operations to consecutive addresses can be
 *      made without reloading EEADR for each byte.
 */

void vEE_Write (char W)
{
#asm
        movwf   EEDATA  ; store data to be programmed
        bsf     RP0     ; RAM bank 1
        bcf     T0IE    ; disable interrupts
        bsf     WREN    ; set EEPROM write enable bit
        movlw   0x55    ; magic lock-up sequence
        movwf   EECON2
        movlw   0xAA
        movwf   EECON2
        bsf     WR      ; start programming
        bsf     T0IE    ; interrupts allowed again
prog_wait:
        btfsc   WR      ; wait for programming operation to finish
        goto    prog_wait

        bcf     RP0     ; RAM bank 0
        incf    EEADR,f ; write pointer++
#endasm
}


/*
 *      Read from EEPROM to RAM
 *
 *      Arguments:
 *        bCount = number of bytes to read.
 *        W = RAM destination address
 *
 *      Before calling vEE_Read(), the caller should load the
 *      appropriate EEPROM start address into PIC register EEADR.
 */

void vEE_Read (char bCount, char W)
{
#asm
        movwf   FSR
ee_read_loop:
        bsf     RP0
        bsf     RD
        bcf     RP0
        movf    EEDATA,w
        movwf   INDF
        incf    EEADR,f
        incf    FSR,f
        decfsz  bCount,f
        goto    ee_read_loop
#endasm
}


/*
 *      Check pushbuttons on RB4 and RB5.
 *
 *      RB5 is debounced and qCalEdge will be set
 *      if a falling edge is detected.
 *
 *      RB4 is simply inverted and copied to qDFCW_Switch.
 */

void vCheckSwitches (void)
{
  TRISB = 0b00111111;       // make RB4 and RB5 inputs

  qDFCW_Switch = !RB4;      // copy this key to a variable for use by main

  if (RB5 == 0)             // CAL switch pressed?
  {
    if (bDeb_RB5 == 1)      // yes, debounce counter goes 1 -> 0 now?
      qCalEdge = TRUE;      // yes, set keypress edge flag (cleared elsewhere)
    if (bDeb_RB5)           // decrement debounce counter down to zero
      bDeb_RB5--;
  }
  else                      // CAL key is released
    bDeb_RB5 = 255;         // reload debounce time

  TRISB = 0b00001111;       // make RB4 and RB5 outputs again (drives LCD)
}


/*
 *      2.6 seconds delay
 */

void vDelay2600 (void)
{
  char bLoop;

  bLoop = 8;
  do
    vDelay (255);
  while (--bLoop);
}


/*
 *      Power-up initialization
 */

void vFirstInit (void)
{
  char bLoop;

  // Set up initial output levels and pin directions of the PIC I/O pins

  PORTA = 0b00000;
  PORTB = 0b10000000;   // AD9832 FSYNC should be idling high
  TRISA = 0b10000;      // all PORTA pins except RA4 are outputs
  TRISB = 0b00001111;   // upper nibble of PORTB are outputs

  // Set up the OPTION register
  //
  // Bit 7 = /RBPU  = 0, PORTB internal pull-up enabled
  // Bit 6 = INTEDG = 0, RB0 interrupt on falling edge (not used)
  // Bit 5 = T0CS   = 0, internal TMR0 clock (Fxtal/4)
  // Bit 4 = T0SE   = 0, rising edge on RA4 increments TMR0 (not used)
  // Bit 3 = PSA    = 1, prescaler assigned to WDT (not TMR0)
  // Bit 2..0 = PS2..PS0 = 111, WDT prescaler = 128 (WDT currently not used)

  clrwdt ();
  OPTION = 0b00001111;

  //    Initialize the LCD (HD44780 controller), 4-bit mode
  //    (E, R/W and RS are low when we get here)

  vDelay (79);       // 101 ms
  bLoop = 3;
  do
  {
    PORTA = 0x03;    // select 8-bit interface three times
    o_lcd_e = 1;     // make E pulse
    o_lcd_e = 0;
    vDelay (25);     // 32 ms
  } while (--bLoop);

  PORTA = 0x02;      // interface is now guaranteed to be 8-bit, set 4-bit
  o_lcd_e = 1;       // make E pulse
  o_lcd_e = 0;
  vDelay (25);       // 32 ms

  vLCD_Ctrl (0x28);  // 4-bit interface, 2 lines, 5x7 font
  vLCD_Ctrl (0x08);  // display off, cursor off, blink off
  vLCD_Ctrl (0x01);  // clear display, home cursor
  vLCD_Ctrl (0x06);  // increment mode, no display shift
  vLCD_Ctrl (0x0C);  // display on, no cursor, no blink

  //    Install default oscillator constant if requested

  TRISB.5 = 1;                  // set RB5 as input
  W = PORTB;                    // read RB5
  TRISB.5 = 0;                  // set RB5 as output
  if ((W & 0x20) == 0)          // is user pressing CAL button at power-up?
  {
    EEADR = eeOSC;              // yes, install default value in EEPROM
    vEE_Write (0x12);
    vEE_Write (0x77);
    vEE_Write (0xCC);
    vEE_Write (0xAB);
    vLCD_Data ('C');
    vLCD_Data ('A');
    vLCD_Data ('L');
    vLCD_Data ('=');
    vLCD_Data ('D');
    vLCD_Data ('e');
    vLCD_Data ('f');
    vLCD_Data ('a');
    vLCD_Data ('u');
    vLCD_Data ('l');
    vLCD_Data ('t');
    vDelay2600 ();
  }

  //    Splash screen

  // 0123456789ABCDEF

  // G0MRF LF DDS VFO
  //   Version 1.00

  vLCD_Ctrl (0x01);  // clear_screen ()
  vLCD_Data ('G');
  vLCD_Data ('0');
  vLCD_Data ('M');
  vLCD_Data ('R');
  vLCD_Data ('F');
  vLCD_Data (32);
  vLCD_Data ('L');
  vLCD_Data ('F');
  vLCD_Data (32);
  vLCD_Data ('D');
  vLCD_Data ('D');
  vLCD_Data ('S');
  vLCD_Data (32);
  vLCD_Data ('V');
  vLCD_Data ('F');
  vLCD_Data ('0');
  vLCD_Ctrl (0xC2);  // gotoxy (2, 1)
  vLCD_Data ('V');
  vLCD_Data ('e');
  vLCD_Data ('r');
  vLCD_Data ('s');
  vLCD_Data ('i');
  vLCD_Data ('o');
  vLCD_Data ('n');
  vLCD_Data (32);
  vLCD_Data ('1');
  vLCD_Data ('.');
  vLCD_Data ('0');
  vLCD_Data ('0');
  vDelay2600 ();

  //    Initialize the AD9832 DDS chip

  vSend_DDS (0xF800);  // SLEEP = RESET = CLR = 1
  vSend_DDS (0xB000);  // SYNC = SELSRC = 1
  vSend_DDS (0xC000);  // SLEEP = RESET = 0

  //    Copy oscillator constant from EEPROM to RAM

  EEADR = eeOSC;        // EEPROM start address for read
  vEE_Read (4, &osc0);  // copy 4 bytes to osc0..osc3

  //    Set start-up frequency

  base_fq0 = 0xA0;      // 0x00085CA0 = 4 * 137000
  base_fq1 = 0x5C;
  base_fq2 = 0x08;
  base_fq3 = 0x00;

  //    Initialize a few other variables

  encstat = PORTB;      // read encoder state to prevent start-up glitch
  qKnobChange = TRUE;   // make main loop update display & DDS once
  bDFWC_Offset = 20;    // 4 * 5 Hz

  //    Enable interrupt

  GIE = 1;
  T0IE = 1;
}


/*
 *      Adjust oscillator constant
 *
 *      Adjustment range is +/- 200 ppm
 *      (+/-5 kHz on the 25 MHz clock)
 */

void vCalibrate (void)
{
  char bOld0, bOld1, bOld2, bOld3;

  vLCD_Ctrl (0x01);        // clear_screen ()
  vLCD_Data ('C');
  vLCD_Data ('A');
  vLCD_Data ('L');
  vLCD_Data (':');
  vLCD_Ctrl (0xC0);        // gotoxy (0, 1)
  vLCD_Data ('0');
  vLCD_Data ('x');
  T0IE = 0;                // atomic copy to knob
#asm                       // use assembler here to save a few program words
        movf    osc0,w
        movwf   bOld0      // bOld = osc (bOld used to check if changed)
        movwf   knob0      // knob = osc
        movf    osc1,w
        movwf   bOld1
        movwf   knob1
        movf    osc2,w
        movwf   bOld2
        movwf   knob2
        movf    osc3,w
        movwf   bOld3
        movwf   knob3
#endasm
  min0 = 0xB8;             // low limit = 2^56 / 25005000 = 0xABC3ABB8
  min1 = 0xAB;
  min2 = 0xC3;
  min3 = 0xAB;
  max0 = 0x52;             // high limit = 2^56 / 24995000 = 0xABD54352
  max1 = 0x43;
  max2 = 0xD5;
  max3 = 0xAB;
  T0IE = 1;
  qKnobChange = TRUE;      // for first update

  do
  {
    vCheckSwitches ();

    if (qKnobChange)            // knob rotated?
    {
      T0IE = 0;                 // atomic copy
      qKnobChange = FALSE;
      osc0 = knob0;             // osc = knob
      osc1 = knob1;
      osc2 = knob2;
      osc3 = knob3;
      T0IE = 1;
      vLCD_Ctrl (0xC2);         // gotoxy (2, 1)
      vWriteHexByte (osc3);     // show on LCD
      vWriteHexByte (osc2);
      vWriteHexByte (osc1);
      vWriteHexByte (osc0);
      vUpdate_DDS ();
    }
  } while (!qCalEdge);          // stay here until user presses CAL again

  if ((bOld0 != osc0) ||        // value changed?
      (bOld1 != osc1) ||
      (bOld2 != osc2) ||
      (bOld3 != osc3))
  {
    EEADR = eeOSC;              // yes, save new value in EEPROM
    vEE_Write (osc0);
    vEE_Write (osc1);
    vEE_Write (osc2);
    vEE_Write (osc3);
  }
  qCalEdge = FALSE;
}


/*
 *      Adjust DFCW offset
 */

void vSet_DFCW_Offset (void)
{
  T0IE = 0;                     // go atomic
  min0 = 0;                     // low limit = 0 Hz
  min1 = 0;
  min2 = 0;
  min3 = 0;
  max0 = 0;
  max1 = 200;                   // high limit = 4 * 50 Hz = 200 Hz
  max2 = 0;
  max3 = 0;
  knob0 = 0;
  knob1 = bDFWC_Offset;         // knob1 = DFCW offset (speed/256)
  knob2 = 0;
  knob3 = 0;
  T0IE = 1;

  do
  {
    bDFWC_Offset = knob1;
    vLCD_Ctrl (0xC1);                 // gotoxy (1, 1)
    vLCD_Data ('=');
    vShow_DFCW_Offs (bDFWC_Offset);
    while (!qKnobChange)
    {
      vCheckSwitches ();
      if (!qDFCW_Switch)              // stay here until DFCW key is released
        return;
    }
    qKnobChange = FALSE;
  } while (TRUE);
}


/*
 *      Initialize LCD & some variables for normal VFO operation
 */

void vInit_VFO_mode (void)
{
  // Write static text on LCD

  vLCD_Ctrl (0x01);             // clear_screen ()
  vLCD_Ctrl (0x80+11);          // gotoxy (11, 0)
  vLCD_Data ('k');
  vLCD_Data ('H');
  vLCD_Data ('z');

  // Show DFCW stuff

  vLCD_Ctrl (0xC0+0);         // gotoxy (0, 1)
  vLCD_Data ('D');
  vLCD_Ctrl (0xC0+2);         // gotoxy (2, 1)
  vShow_DFCW_Offs (bDFWC_Offset);
  vLCD_Ctrl (0xC0+8);         // gotoxy (8, 1)
  vLCD_Data ('H');
  vLCD_Data ('z');

  // Copy frequency from base_fq to knob

  T0IE = 0;                     // atomic!
  knob0 = base_fq0;
  knob1 = base_fq1;
  knob2 = base_fq2;
  knob3 = base_fq3;
  T0IE = 1;

  // Set frequency limits

  min0 = 0x50;   // minimum frequency
  min1 = 0x48;   // 0x00084850 = 4 * 135700 Hz
  min2 = 0x08;
  min3 = 0x00;

  max0 = 0x20;   // maximum frequency
  max1 = 0x69;   // 0x00086920 = 4 * 137800 Hz
  max2 = 0x08;
  max3 = 0x00;

  qKnobChange = TRUE;   // force first update of LCD frequency and DDS
  qLastLock = !i_lock;  // force first update of lock indicator
  qLastDFCW = !i_dfcw;  // force first update of DFCW state
}


void main (void)
{
  vFirstInit ();
  vInit_VFO_mode ();

  for (;;)
  {
    vCheckSwitches ();

    if (i_lock != qLastLock)    // change on lock input?
    {
      qLastLock = !qLastLock;   // yes, toggle this
      vLCD_Ctrl (0xCF);         // gotoxy (15, 1)
      if (qLastLock)
        vLCD_Data ('L');        // yes, show 'L' for "Locked"
      else
        vLCD_Data (32);         // no, replace 'L' with a space
    }

    if (qDFCW_Switch)
    {
      vSet_DFCW_Offset ();
      vInit_VFO_mode ();
    }

    if (qCalEdge)
    {
      qCalEdge = FALSE;
      vCalibrate ();
      vInit_VFO_mode ();
    }

    if (qKnobChange)
    {
      T0IE = 0;         // the following copy operation must be atomic
      qKnobChange = FALSE;
      base_fq0 = knob0;
      base_fq1 = knob1;
      base_fq2 = knob2;
      base_fq3 = knob3;
      qNewFreq = TRUE;
      T0IE = 1;         // enable timer tick interrupt again

    }

    if (i_dfcw != qLastDFCW)    // DFCW input changed state?
    {
      qLastDFCW = !qLastDFCW;   // remember for next transition
      qNewFreq = TRUE;
    }

    fq0 = base_fq0;             // base frequency
    fq1 = base_fq1;
    fq2 = base_fq2;
    fq3 = base_fq3;

    if (!qLastDFCW)             // DFCW input low?
    {
#asm
      movf    bDFWC_Offset,w    // yes, add DFCW offset
      addwf   fq0,f
      btfss   Carry
      goto    offs_done
      incfsz  fq1,f
      goto    offs_done
      incfsz  fq2,f
      goto    offs_done
      incf    fq3,f
offs_done:
#endasm
    }

    if (qNewFreq)               // fq changed by knob or DFCW?
    {
      qNewFreq = FALSE;
      vUpdate_DDS ();
      vLCD_Ctrl (0x80);         // gotoxy (0, 0)
      vShowFrequency ();
      vLCD_Ctrl (0xC1);         // gotoxy (1, 1)
      if (!qLastDFCW)           // DFCW offset active?
        vLCD_Data ('+');        // yes, show + sign
      else
        vLCD_Data (32);         // no, replace + with a space
    }
  }
}
